import numpy as np
import os
from limix_lsf import clusterrun
from limix_util import pickle as pickle_
from limix_util.report import BeginEnd, ProgressBar
from limix_util.time import Timer
from limix_util.path import folder_hash
from hcache import Cached, cached

class Job(Cached):
    def __init__(self, workspace_id, experiment_id, jobid):
        super(Job, self).__init__()
        self.jobid = int(jobid)
        self._workspace_id = workspace_id
        self._experiment_id = experiment_id
        self.finished = False
        self.submitted = False
        self.bjobid = None
        self.brunid = None

    @property
    def failed(self):
        if self.submitted:
            bjob = self.get_bjob()
            stats = bjob.exit_status()
            return (stats is not None and stats != 0)
        return False


    @cached
    def get_bjob(self):
        bjob = clusterrun.get_bjob(self.brunid, self.bjobid)
        return bjob

    @property
    def task_ids(self):
        from . import workspace

        e = workspace.get_experiment(self._workspace_id, self._experiment_id)

        task2job = np.floor(e.njobs * np.arange(e.ntasks) / e.ntasks)
        task2job = np.asarray(task2job, int)
        task_ids = list(np.where(task2job == self.jobid)[0])

        return task_ids

    def get_tasks(self):
        from . import workspace

        task_ids = self.task_ids

        e = workspace.get_experiment(self._workspace_id, self._experiment_id)

        tasks = e.get_tasks()
        return [tasks[tid] for tid in task_ids]

    def run(self, progress):
        tasks = self.get_tasks()
        task_results = []

        if progress:
            p = ProgressBar(len(tasks))

        for (i, task) in enumerate(tasks):
            with Timer() as timer:
                tr = task.run()
            tr.total_elapsed = timer.elapsed
            if progress:
                p.update(i+1)
            task_results.append(tr)

        if progress:
            p.finish()

        self.finished = True

        return task_results

    def __str__(self):
        from tabulate import tabulate
        table = [['Job ID', str(self.jobid)]]
        table.append(['Submitted', str(self.submitted)])
        table.append(['Finished', str(self.finished)])
        table.append(['# tasks', len(self.task_ids)])
        table.append(['Task IDs', str(self.task_ids)])

        bjob_status = 'N/A'
        bjob_exit_status = 'N/A'
        bjob_os_id = 'N/A'

        if self.submitted:
            bjob_status = self.get_bjob().stat()
            bjob_exit_status = str(self.get_bjob().exit_status())
            bjob_os_id = self.get_bjob().os_jobid

        table.append(['Bjob status', bjob_status])
        table.append(['Bjob exit status', bjob_exit_status])
        table.append(['Bjob OS id', bjob_os_id])

        return tabulate(table)

def store_jobs(jobs, fpath):
    with BeginEnd('Storing jobs'):
        pickle_.pickle({t.jobid:t for t in jobs}, fpath)
        print('   %d jobs stored   ' % len(jobs))

def store_job(job, fpath):
    pickle_.pickle(job, fpath)

def load_job(fpath):
    return pickle_.unpickle(fpath)

def collect_jobs(folder):
    exist = os.path.exists(os.path.join(folder, 'all.pkl'))

    if exist:
        ha = folder_hash(folder, ['all.pkl', '.folder_hash'])
        fh = os.path.join(folder, '.folder_hash')
        ok = os.path.exists(fh)
        ok = ok and ha == open(os.path.join(folder, '.folder_hash')).read(32)
        if ok:
            with BeginEnd('Unpickling jobs'):
                return pickle_.unpickle(os.path.join(folder, 'all.pkl'))

    return pickle_.pickle_merge(folder)




from argparse import ArgumentParser
import re
import inspect
import os
from os.path import join
import json
import imp
import shutil
from .config import conf
from . import experiment
from limix_util.report import BeginEnd
from limix_util.inspect import fetch_classes, fetch_functions
import limix_lsf
import logging

_workspaces = dict()
def get_workspace(workspace_id):
    if workspace_id not in _workspaces:
        _workspaces[workspace_id] = Workspace(workspace_id)
    return _workspaces[workspace_id]

def get_experiment(workspace_id, experiment_id):
    w = get_workspace(workspace_id)
    if w is None:
        raise Exception('There is no workspace called %s.' % workspace_id)

    e = w.get_experiment(experiment_id)
    if e is None:
        raise Exception('There is no experiment '
                        'called %s in the workspace %s.'
                        % (experiment_id, workspace_id))

    return e

def exists(workspace_id):
    folder = join(conf.get('default', 'base_dir'), workspace_id)
    return os.path.exists(folder)

def get_workspace_ids():
    files = os.listdir(conf.get('default', 'base_dir'))
    return [f for f in files if os.path.isdir(f)]

class Workspace(object):
    def __init__(self, workspace_id):
        super(Workspace, self).__init__()
        self._workspace_id = workspace_id
        self._script_filepath = None
        self._experiments = dict()
        self.force_cache = False
        self._logger = logging.getLogger(__name__)
        self._logger.debug('Workspace %s has been created.', workspace_id)

    def rm_experiment(self, experiment_id):
        e = self.get_experiment(experiment_id)
        e.kill_bjobs()
        if os.path.exists(e.folder):
            with BeginEnd('Removing folder'):
                shutil.rmtree(e.folder)

    def get_properties(self):
        try:
            with open(join(self.folder, 'properties.json')) as json_file:
                return json.load(json_file)
        except ValueError as e:
            print(e)
            print("File: %s", join(self.folder, 'properties.json'))
            print('File content:')
            print(open(join(self.folder, 'properties.json')).read())
            os.system("hostname")

    def get_plot_class(self, name):
        return self._get_plot_classes_map()[name]

    def _get_plot_classes_map(self):
        f = join(self.folder, 'plot.json')
        self._logger.info("Reading plot config file %s.", f)
        with open(f) as json_file:
            fpaths = json.load(json_file)
            fpaths = [str(fpath) for fpath in fpaths]

        plot_classes_map = dict()
        for fp in fpaths:
            self._logger.info("Reading file %s.", fp)
            clss = fetch_classes(fp, '.*Plot')
            for clsi in clss:
                plot_classes_map[clsi.__name__] = clsi
        return plot_classes_map

    @property
    def folder(self):
        return join(conf.get('default', 'base_dir'), self._workspace_id)

    def _get_auto_run(self, experiment_id):
        auto_runs_map = self._get_auto_runs_map()
        # if experiment_id not in auto_runs_map:
            # return None
        return auto_runs_map[experiment_id]

    def get_experiment(self, experiment_id):
        if experiment_id not in self._experiments:
            self._setup_experiment(experiment_id)
        return self._experiments[experiment_id]

    def _setup_experiment(self, experiment_id):
        self._experiments[experiment_id] =\
            experiment.Experiment(self._workspace_id, experiment_id,
                                  self.get_properties())
        auto_run = self._get_auto_run(experiment_id)
        if auto_run is None:
            return
        auto_run(self._experiments[experiment_id])
        self._experiments[experiment_id].auto_run_done = True
        self._experiments[experiment_id].finish_setup()

    def _get_auto_runs_map(self):
        fps = self._auto_run_filepaths()
        auto_runs_map = dict()
        for fp in fps:
            self._logger.info("Reading file %s.", fp)
            funcs = fetch_functions(fp, r'^auto_run_.+$')
            auto_runs_map.update([(f.__name__[9:], f) for f in funcs])
        return auto_runs_map

    @property
    def dataset_folder(self):
        return os.path.join(self.folder, 'dataset')

    def _auto_run_filepaths(self):
        f = join(self.folder, 'auto_run.json')
        if not os.path.exists(f):
            self._logger.warn("File %s does not exist.", f)
            return []
        with open(f) as json_file:
            fpaths = json.load(json_file)
            fpaths = [str(fpath) for fpath in fpaths]
            return fpaths

    def auto_run(self, args):
        filepaths = self._auto_run_filepaths()
        for fp in filepaths:
            self._auto_run(fp, args)

    def _auto_run(self, script_filepath, args):

        p = ArgumentParser()
        p.add_argument('--dryrun', dest='dryrun', action='store_true')
        p.add_argument('--no-dryrun', dest='dryrun', action='store_false')
        p.add_argument('--queue', default=None)
        p.set_defaults(dryrun=False, parallel=True)

        args = p.parse_args(args)

        lista = self._get_generate_tasks(script_filepath)
        for (experiment_id, func) in lista:
            exp = self.get_experiment(experiment_id)
            func(exp)

            try:
                exp.submit_jobs(args.dryrun, queue=args.queue)
            except KeyboardInterrupt:
                shutil.rmtree(exp.folder)
                raise

            if args.dryrun:
                shutil.rmtree(exp.folder)

    def remove(self, experiment_id, jobs_too=False):
        if experiment_id is None:
            print('Nothing to remove.')
            return
        if len(experiment_id) == 0:
            print('Nothing to remove.')
            return
        with BeginEnd('Removing folder'):
            shutil.rmtree(join(self.folder, experiment_id))

        if jobs_too:
            bgroup = '/%s/%s' % (self._workspace_id, experiment_id)
            if exists(bgroup):
                limix_lsf.util.kill_group(bgroup, True)

    def _call_job(self, args):
        experiment_id = args.experiment_id
        (experiment_id, auto_run_func) = self._get_generate_tasks(experiment_id)
        exp = self.get_experiment(experiment_id)
        exp.script_filepath = self._script_filepath
        auto_run_func(exp)
        exp.start(check_existence=False)

    def _get_generate_tasks(self, script_filepath, experiment_id=None):
        script_name = os.path.basename(script_filepath)
        mod_name = os.path.splitext(script_name)[0]
        mod = imp.load_source(mod_name, script_filepath)

        lista = []
        for func in inspect.getmembers(mod, inspect.isfunction):
            func_name = func[0]
            m = re.match(r'^auto_run_(.+)$', func_name)
            if m:
                exp_name = m.group(1)
                if experiment_id is None:
                    lista.append((exp_name, func[1]))
                elif exp_name == experiment_id:
                    return exp_name, func[1]

        return lista

    def __str__(self):
        from tabulate import tabulate

        files = os.listdir(self.folder)
        exp_ids = [f for f in files if (os.path.isdir(join(self.folder, f)) and
                                        join(self.folder, f,
                                             'generate_tasks.txt'))]

        auto_run_err = set()
        tasks_setup_err = set()
        finish_setup_err = set()
        imp_err = set()
        suc = set()

        for ei in exp_ids:
            try:
                e = self.get_experiment(ei)
                if not e.auto_run_done:
                    auto_run_err.add(ei)
                elif not e.tasks_setup_done:
                    tasks_setup_err.add(ei)
                elif not e.finish_setup_done:
                    finish_setup_err.add(ei)
                else:
                    suc.add(ei)
            except ImportError:
                imp_err.add(ei)

        table = []
        auto_run_fps = '\n'.join(self._auto_run_filepaths())
        table.append(['auto_run files', auto_run_fps])
        table.append(['# experiments', str(len(exp_ids))])


        table.append(["# auto_run err", len(auto_run_err)])
        table.append(["# tasks setup err", len(tasks_setup_err)])
        table.append(["# finish setup err", len(finish_setup_err)])
        table.append(["# import err", len(imp_err)])
        table.append(["# successful", len(suc)])

        msg = tabulate(table)

        return msg

def _get_auto_runs_map(script_filepath):
    script_name = os.path.basename(script_filepath)
    mod_name = os.path.splitext(script_name)[0]
    mod = imp.load_source(mod_name, script_filepath)

    auto_run_map = dict()
    for func in inspect.getmembers(mod, inspect.isfunction):
        func_name = func[0]
        m = re.match(r'^auto_run_(.+)$', func_name)
        if m:
            exp_name = m.group(1)
            auto_run_map.update({exp_name:func[1]})

    return auto_run_map



from limix_exp.base import BasePlot
from limix_tool.qtl import NullScore
from limix_tool.qtl import combine_pvalues
# from gwarped_exp import workspace
from limix_plot.qqplot import QQPlot

class CalPlot(BasePlot):
    def __init__(self, tasks, group_by=None):
        super(CalPlot, self).__init__(tasks, group_by)

        if len(tasks) > 0:
            self._workspace = workspace.get_workspace(tasks[0]._workspace_id)
        else:
            self._workspace = None

    def _plot_tasks(self, tasks, ax):
        w = workspace.get_workspace(tasks[0]._workspace_id)
        properties = w.get_properties()

        methods = tasks[0].get_result().methods

        nfails = {m:0 for m in methods}

        ns = NullScore()

        for method in methods:
            pvs = [task.get_result().pv(method) for task in tasks]
            pv = combine_pvalues(pvs)
            ns.add(method, pv)

        qqplot = QQPlot()
        for method in methods:
            qqplot.add(ns.pv(method), method)
            confid = ns.confidence_band(method)
            qqplot.set_confidence(confid, method)

        qqplot.plot_top = 10.
        qqplot.plot(ax)
        ax.set_xlim([0, 20])
        ax.set_ylim([0, 20])




import os
from os.path import join
import random
import logging
from humanfriendly import format_size
from humanfriendly import parse_size
from tabulate import tabulate
from limix_lsf import clusterrun
from limix_lsf.clusterrun import ClusterRun
from limix_util import path as path_
from limix_util.path import make_sure_path_exists
from limix_util.report import BeginEnd, ProgressBar
from hcache import Cached, cached
from . import task
from . import job
from .config import conf

class Experiment(Cached):
    def __init__(self, workspace_id, experiment_id, properties):
        super(Experiment, self).__init__()
        self._workspace_id = workspace_id
        self._experiment_id = experiment_id

        self.script_filepath = None
        self._task_id_counter = -1
        self.njobs = None
        self.mkl_nthreads = 1
        self.nprocs = 1
        self._job_megabytes = None
        self._properties = properties
        self.auto_run_done = False
        self.finish_setup_done = False
        self._logger = logging.getLogger(__name__)
        self._logger.debug('Experiment %s has been created.', experiment_id)

    @property
    def runid(self):
        fp = join(self.folder, '.runid')
        if os.path.exists(fp):
            return open(fp, 'r').read()

    @runid.setter
    def runid(self, v):
        fp = join(self.folder, '.runid')
        open(fp, 'w').write(v)

    def kill_bjobs(self):
        jobs = self.get_jobs()
        runids = set([j.brunid for j in jobs if j.submitted])
        for ri in runids:
            if clusterrun.exists(ri):
                clusterrun.load(ri).kill()
                clusterrun.rm(ri)
            else:
                self._logger.warn('Cluster run %s does not exist.' % ri)

    @property
    def job_memory(self):
        nbytes = int(round(self._job_megabytes * 1024. * 1024.))
        return format_size(nbytes)

    @job_memory.setter
    def job_memory(self, siz):
        nbytes = parse_size(siz)
        self._job_megabytes = int(round(nbytes / 1024. / 1024.))

    def exists(self):
        folder = self.folder
        return os.path.exists(folder)

    def create_task(self):
        self._task_id_counter += 1
        return task.Task(self._workspace_id, self._experiment_id,
                         self._task_id_counter)

    def get_task(self, task_id):
        return self._get_tasks()[task_id]

    @cached
    def _get_task_results(self):
        fpath = join(self.folder, 'result')
        return task.collect_task_results(fpath, force_cache=False)

    @cached
    def get_task_results(self):
        return list(self._get_task_results().values())

    def get_task_result(self, task_id):
        return self._get_task_results().get(task_id)

    def has_task_result(self, task_id):
        return int(task_id) in self._get_task_results()

    @property
    def folder(self):
        return join(conf.get('default', 'base_dir'),
                    self._workspace_id,
                    self._experiment_id)

    @property
    def figures_folder(self):
        f = join(self.folder, 'figs')
        make_sure_path_exists(f)
        return f

    def split_folder(self, jobid):
        return str(int(jobid/1000))

    @cached
    def _get_tasks(self):
        fpath = join(self.folder, 'tasks.pkl')
        return task.load_tasks(fpath)

    def get_tasks(self):
        return list(self._get_tasks().values())

    @property
    def ntasks(self):
        return len(self.get_tasks())

    def get_task_args(self):
        fpath = join(self.folder, 'task_args.pkl')
        return task.load_task_args(fpath)

    def define_task_args(self, task_args):
        raise NotImplementedError

    def generate_tasks(self):
        raise NotImplementedError

    def generate_jobs(self, workspace_id):
        wid = workspace_id
        eid = self._experiment_id
        jobs = [job.Job(wid, eid, i) for i in range(self.njobs)]
        return jobs

    def do_task(self, task):
        raise NotImplementedError

    @property
    def tasks_setup_done(self):
        fpath = join(self.folder, 'tasks.pkl')
        return os.path.exists(fpath)

    def _store_tasks(self, tasks):
        fpath = join(self.folder, 'tasks.pkl')
        task.store_tasks(tasks, fpath)

    def _store_jobs(self, jobs):
        with BeginEnd('Storing jobs'):
            pb = ProgressBar(len(jobs))
            for (i, j) in enumerate(jobs):
                fp = self.job_path(j.jobid)
                make_sure_path_exists(os.path.dirname(fp))
                job.store_job(j, fp)
                pb.update(i+1)
            pb.finish()

    def job_path(self, jobid):
        fp = join(self.folder, 'job', self.split_folder(jobid))
        fp = join(fp, str(jobid) + '.pkl')
        return fp

    def task_result_path(self, jobid):
        fp = join(self.folder, 'result', self.split_folder(jobid))
        fp = join(fp, str(jobid) + '.pkl')
        return fp

    def _store_task_args(self, task_args):
        fpath = join(self.folder, 'task_args.pkl')
        task.store_task_args(task_args, fpath)

    def run_job(self, jobid, progress=True, dryrun=False, force=False):
        job_ = job.load_job(self.job_path(jobid))

        if job_.finished and not force:
            print("Job %d has already finished." % jobid)
            return

        task_results = job_.run(progress)

        if not dryrun:
            job.store_job(job_, self.job_path(job_.jobid))
            fp = self.task_result_path(job_.jobid)
            make_sure_path_exists(os.path.dirname(fp))
            task.store_task_results(task_results, fp)

    @property
    def are_init_jobs_files_generated(self):
        fp = join(self.folder, '.init_jobs_files_generated')
        return os.path.exists(fp)

    def finish_setup(self):
        make_sure_path_exists(self.folder)

        ta = task.TaskArgs()
        self.define_task_args(ta)
        self._store_task_args(ta)

        if self.tasks_setup_done:
            tasks = self.get_tasks()
        else:
            with BeginEnd('Generating tasks'):
                tasks = self.generate_tasks()
            self._store_tasks(tasks)

        ntasks = len(tasks)
        if self.njobs is None:
            self.njobs = ntasks
        else:
            self.njobs = min(self.njobs, ntasks)

        if not self.are_init_jobs_files_generated:
            with BeginEnd('Generating jobs'):
                jobs = self.generate_jobs(self._workspace_id)
                if len(jobs) == 0:
                    raise Exception('No job has been generated.')
                print('   %d generated jobs   ' % len(jobs))
            self._store_jobs(jobs)
            fp = join(self.folder, '.init_jobs_files_generated')
            path_.touch(fp)

        self.finish_setup_done = True

    @cached
    def get_job(self, jobid):
        return job.load_job(self.job_path(jobid))

    @cached
    def get_jobs(self):
        folder = join(self.folder, 'job')
        jobs = job.collect_jobs(folder)
        keys = list(jobs.keys())
        vals = list(jobs.values())
        return [j for (_, j) in sorted(zip(keys, vals))]

    def _get_job_task_ids(self, jobid):
        import numpy as np
        ntasks = len(self.get_tasks())
        task2job = np.floor(self.njobs * np.arange(ntasks) / ntasks)
        task2job = np.asarray(task2job, int)
        task_ids = list(np.where(task2job == jobid)[0])
        return task_ids

    def resubmit(self, jobid):
        job = self.get_job(jobid)
        cr = clusterrun.load(self.runid)
        cr.resubmit(job.bjobid)

    def submit_jobs(self, dryrun, requests=None, queue=None, verbose=False):
        jobs = self.get_jobs()
        myrand = random.Random(937628)
        myrand.shuffle(jobs)

        title = '/%s/%s' % (self._workspace_id, self._experiment_id)
        cmd = ClusterRun(title)
        cmd.queue = queue
        cmd.memory = self.job_memory
        cmd.mkl_nthreads = self.mkl_nthreads
        cmd.nprocs = self.nprocs
        if requests is not None:
            for request in requests:
                cmd.request(request)

        for j in jobs:
            a = ['arauto']
            if self._logger.isEnabledFor(logging.DEBUG):
                a += ['--verbose']
            a += ['rjob', self._workspace_id]
            a += [self._experiment_id]
            a += [j.jobid]
            if dryrun:
                a += ['--dryrun']
            else:
                a += ['--no-dryrun']
            a += ['--no-progress']
            cmd.add(a)

        self.runid = cmd.run(dryrun=dryrun)
        for (i, j) in enumerate(cmd.jobs):
            jobs[i].bjobid = j.jobid
            jobs[i].brunid = j.runid
            jobs[i].submitted = True

        if not dryrun:
            self._store_jobs(jobs)
            cmd.store()

    def method_errors(self):
        tasks = self.get_tasks()
        if len(tasks) == 0:
            return
        task_results = self.get_task_results()
        properties = self._properties

        methods = task_results[0].methods
        err_msgs = {m:set() for m in methods}
        for tr in task_results:
            for method in methods:
                s = err_msgs[method]
                if tr.error_status != 0:
                    s.add(tr.error_msg(method))

        ml = {m:properties[m]['label'] for m in methods}

        for m in methods:
            print('Error messages for %s:' % ml[m])
            if len(err_msgs[m]) > 0:
                for msg in err_msgs[m]:
                    print(msg)

    def _store_task_results(self, folder_split, jobid, task_results):
        base = join(self.folder, 'result', folder_split)
        make_sure_path_exists(base)
        fpath = join(base, '%d.pkl' % int(jobid))
        task.store_task_results(task_results, fpath)

    @property
    def bgroup(self):
        return '/' + self._workspace_id + '/' + self._experiment_id

    def __str__(self):
        with BeginEnd('Compiling info table'):
            table = []
            table.append(['# jobs', str(self.njobs)])
            table.append(['# tasks', str(self.ntasks)])

            bjobs_finished = [j.get_bjob() for j in self.get_jobs() if j.submitted and j.finished]

            nsub = sum([1 for j in self.get_jobs() if j.submitted])
            table.append(['# submitted jobs', str(nsub)])

            npend = sum([1 for j in self.get_jobs() if (j.submitted and
                                                        j.get_bjob().ispending()
                                                       )])
            table.append(['# pending jobs', str(npend)])

            nrun = sum([1 for j in self.get_jobs() if (j.submitted and
                                                       j.get_bjob().isrunning()
                                                      )])
            table.append(['# running jobs', str(nrun)])

            nfin = sum([1 for j in self.get_jobs() if j.finished])
            table.append(['# finished jobs', str(nfin)])

            failed_jobids = [j.jobid for j in self.get_jobs() if j.failed]
            nfail = len(failed_jobids)
            table.append(['# failed jobs', str(nfail)])

            resource_info = [bj.resource_info() for bj in bjobs_finished]

            max_memories = [r['max_memory'] for r in resource_info if r is not None and r['max_memory'] is not None]
            if len(max_memories) > 0:
                max_memory = max(max_memories)
            else:
                max_memory = None
            req_memories = [r['req_memory'] for r in resource_info if r is not None and r['req_memory'] is not None]
            if len(req_memories) > 0:
                req_memory = max(req_memories)
            else:
                req_memory = None
            table.append(['max used memory', format_size(max_memory) if max_memory is not None else 'n/a'])
            table.append(['max req. memory', format_size(req_memory) if req_memory is not None else 'n/a'])

            msg = tabulate(table)

            if nfail > 0:
                msg += '\nFailed jobs: ' + str(failed_jobids)
                msg += '\n'

        return msg


def _namespace2task(workspace_id, experiment_id, n):
    n = vars(n)
    t = task.Task(workspace_id, experiment_id, n['task_id'])
    for (k, v) in list(n.items()):
        setattr(t, k, v)
    return t




import os
from tabulate import tabulate
from limix_util.pickle import SlotPickleMixin
from limix_util.pickle import PickleByName
from limix_util import pickle
from limix_util.path import folder_hash
from limix_util.report import BeginEnd
from limix_util.scalar import isfloat
from limix_util.string import summarize

def extract_successes_and_failures(tasks):
    methods = tasks[0].get_result().methods

    nsucs = {m:0 for m in methods}
    nfails = {m:0 for m in methods}

    for task in tasks:
        tr = task.get_result()
        for m in methods:
            if tr.error_status(m) == 0:
                nsucs[m] += 1
            else:
                nfails[m] += 1

    return dict(nsucs=nsucs, nfails=nfails)

class TaskArgs(SlotPickleMixin):
    def __init__(self):
        super(TaskArgs, self).__init__()
        self._names = []

    def add(self, name):
        self._names.append(name)

    def get_names(self):
        return self._names

class Task(PickleByName):
    def __init__(self, workspace_id, experiment_id, task_id):
        super(Task, self).__init__()
        self.task_id = int(task_id)
        self.workspace_id = workspace_id
        self.experiment_id = experiment_id

    def run(self):
        from .workspace import get_experiment
        e = get_experiment(self.workspace_id, self.experiment_id)
        return e.do_task(self)

    @property
    def finished(self):
        return self.get_result() is not None

    def get_result(self):
        from .workspace import get_experiment
        e = get_experiment(self.workspace_id, self.experiment_id)
        if e is None:
            return None
        return e.get_task_result(self.task_id)

class TaskResult(SlotPickleMixin):
    __slots__ = ['total_elapsed', 'workspace_id', 'experiment_id', 'task_id',
                 '_elapsed', '_error_status', '_error_msg', '_methods']

    def __init__(self, workspace_id, experiment_id, task_id):
        super(TaskResult, self).__init__()
        self.total_elapsed = float('nan')
        self.workspace_id = workspace_id
        self.experiment_id = experiment_id
        self.task_id = int(task_id)
        self._elapsed = dict()
        self._error_status = dict()
        self._error_msg = dict()
        self._methods = set()

    def get_task(self):
        from .workspace import get_experiment
        e = get_experiment(self.workspace_id, self.experiment_id)
        return e.get_task(self.task_id)

    def elapsed(self, method):
        return self._elapsed[method]

    def error_status(self, method):
        return self._error_status[method]

    def error_msg(self, method):
        return self._error_msg[method]

    @property
    def methods(self):
        return list(self._methods)

    def set_error_status(self, method, error_status):
        self._add_method(method)
        self._error_status[method] = int(error_status)

    def set_error_msg(self, method, error_msg):
        self._add_method(method)
        self._error_msg[method] = str(error_msg)

    def set_elapsed(self, method, elapsed):
        self._add_method(method)
        self._elapsed[method] = float(elapsed)

    def _add_method(self, method):
        self._methods.add(method)

def load_tasks(fpath, verbose=False):
    with BeginEnd('Loading tasks', silent=not verbose):
        if os.path.exists(fpath):
            if verbose:
                print("Exist %s" % fpath)
        else:
            print("Does not exist %s" % fpath)
        tasks = pickle_.unpickle(fpath)
        if verbose:
            print('   %d tasks found  ' % len(tasks))
    return tasks

def store_tasks(tasks, fpath):
    if os.path.exists(fpath):
        return
    pickle_.pickle({t.task_id:t for t in tasks}, fpath)

def load_task_args(fpath):
    return pickle_.unpickle(fpath)

def store_task_args(task_args, fpath):
    if os.path.exists(fpath):
        return
    pickle_.pickle(task_args, fpath)

def collect_task_results(folder, force_cache=False):
    assert force_cache is False
    fpath = os.path.join(folder, 'all.pkl')
    exist = os.path.exists(fpath)

    if exist:
        ha = folder_hash(folder, ['all.pkl', '.folder_hash'])
        fh = os.path.join(folder, '.folder_hash')
        ok = os.path.exists(fh)
        ok = ok and ha == open(os.path.join(folder, '.folder_hash')).read(32)
        if ok:
            with BeginEnd('Unpickling tasks'):
                return pickle_.unpickle(os.path.join(folder, 'all.pkl'))

    return pickle_.pickle_merge(folder)

def store_task_results(task_results, fpath):
    with BeginEnd('Storing task results'):
        pickle_.pickle({tr.task_id:tr for tr in task_results}, fpath)
        print('   %d task results stored   ' % len(task_results))

def tasks_summary(tasks):
    from collections import OrderedDict
    from .workspace import get_experiment

    if len(tasks) == 0:
        return ''

    wid = tasks[0].workspace_id
    eid = tasks[0].experiment_id

    e = get_experiment(wid, eid)

    args = e.get_task_args().get_names()

    args.sort()
    d = OrderedDict([(k, set()) for k in args])

    for task in tasks:
        for a in args:
            d[a].add(getattr(task, a))

    for a in args:
        d[a] = list(d[a])
        d[a].sort(key= lambda x: float(x) if isfloat(x) else x)

    table = list(zip(list(d.keys()), [summarize(v) for v in list(d.values())]))
    return '*** Task summary ***\n' + tabulate(table)
